/*******************************************************************************
 *
 * htmltokenizer.c - HTML parsing engine
 *
 * Hash functions for identifying html tags.
 *
 * Cheetah Web Browser
 * Copyright (C) 2001 Garett Spencley 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 *
 *******************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "htmltokenizer.h"
#include "html.h"
#include "str_util.h"

/*
 * tag_hash()
 *
 * hashes html tags
 */

__inline static unsigned int tag_hash(const char *str, unsigned int len)
{
	static unsigned short asso_values[] = {
		420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
		420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
		420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
		420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
		420, 420, 420, 420, 420, 420, 420, 0, 420, 35,
		115, 15, 20, 40, 60, 420, 420, 420, 420, 420,
		420, 420, 420, 420, 420, 155, 110, 195, 65, 35,
		35, 85, 44, 209, 420, 15, 105, 230, 20, 170,
		175, 0, 30, 95, 160, 75, 220, 420, 0, 25,
		420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
		420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
		420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
		420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
		420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
		420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
		420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
		420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
		420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
		420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
		420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
		420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
		420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
		420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
		420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
		420, 420, 420, 420, 420, 420, 420, 420, 420, 420,
		420, 420, 420, 420, 420, 420
	};

	return len + asso_values[(unsigned char) str[len - 1]] + asso_values[(unsigned char) str[0]];
}

/*
 * lookup_tag()
 * 
 * Returns the structure for the tag we are looking for
 */

__inline html_token *lookup_tag(const char *str, unsigned int len)
{
	static html_token wordlist[] = {
		{"Q", Q, NULL},
		{"/Q", EO_Q, NULL},
		{"/H3", EO_H3, NULL},
		{"/H4", EO_H4, NULL},
		{"/DFN", EO_DFN, NULL},
		{"/SPAN", EO_SPAN, NULL},
		{"/BUTTON", EO_BUTTON, NULL},
		{"/OPTION", EO_OPTION, NULL},
		{"/CAPTION", EO_CAPTION, NULL},
		{"/BODY", EO_BODY, NULL},
		{"/TBODY", EO_TBODY, NULL},
		{"/TR", EO_TR, NULL},
		{"/DIR", EO_DIRECTORY, NULL},
		{"/VAR", EO_VAR, NULL},
		{"/ABBR", EO_ABBR, NULL},
		{"/NOBR", EO_NOBR, NULL},
		{"/CENTER", EO_CENTER, NULL},
		{"/H1", EO_H1, NULL},
		{"/PRE", EO_PRE, NULL},
		{"/CODE", EO_CODE, NULL},
		{"/BASE", EO_BASE, NULL},
		{"/CITE", EO_CITE, NULL},
		{"/TITLE", EO_TITLE, NULL},
		{"/STYLE", EO_STYLE, NULL},
		{"/FRAME", EO_FRAME, NULL},
		{"/TABLE", EO_TABLE, NULL},
		{"/IFRAME", EO_IFRAME, NULL},
		{"/H5", EO_H5, NULL},
		{"/BLOCKQUOTE", EO_BLOCKQUOTE, NULL},
		{"/TH", EO_TH, NULL},
		{"NOBR", NOBR, NULL},
		{"H3", H3, NULL},
		{"/H6", EO_H6, NULL},
		{"H4", H4, NULL},
		{"/TD", EO_TD, NULL},
		{"/DD", EO_DD, NULL},
		{"/HEAD", EO_HEAD, NULL},
		{"/THEAD", EO_THEAD, NULL},
		{"/LEGEND", EO_LEGEND, NULL},
		{"FRAME", FRAME, NULL},
		{"HR", HR, NULL},
		{"/U", EO_UNDERLINE, NULL},
		{"/MENU", EO_MENU, NULL},
		{"H1", H1, NULL},
		{"KBD", KBD, NULL},
		{"H5", H5, NULL},
		{"DFN", DFN, NULL},
		{"/BIG", EO_BIG, NULL},
		{"/STRONG", EO_STRONG, NULL},
		{"/S", EO_S, NULL},
		{"DIR", DIRECTORY, NULL},
		{"/INS", EO_INS, NULL},
		{"/ADDRESS", EO_ADDRESS, NULL},
		{"/NOFRAMES", EO_NOFRAMES, NULL},
		{"H6", H6, NULL},
		{"/UL", EO_UL, NULL},
		{"/DL", EO_DL, NULL},
		{"/OL", EO_OL, NULL},
		{"/DEL", EO_DEL, NULL},
		{"/HTML", EO_HTML, NULL},
		{"/LABEL", EO_LABEL, NULL},
		{"/SMALL", EO_SMALL, NULL},
		{"/B", EO_BOLD, NULL},
		{"/B", EO_BOLD, NULL},
		{"HEAD", HEAD, NULL},
		{"/KDB", EO_KBD, NULL},
		{"/SUB", EO_SUB, NULL},
		{"/H2", EO_H2, NULL},
		{"SPAN", SPAN, NULL},
		{"NOFRAMES", NOFRAMES, NULL},
		{"LINK", LINK, NULL},
		{"DD", DD, NULL},
		{"STYLE", STYLE, NULL},
		{"BUTTON", BUTTON, NULL},
		{"BODY", BODY, html_body},
		{"BR", BR, html_br},
		{"BASE", BASE, NULL},
		{"U", UNDERLINE, NULL},
		{"HTML", HTML, NULL},
		{"BLOCKQUOTE", BLOCKQUOTE, NULL},
		{"/A", EO_A, NULL},
		{"/AREA", EO_AREA, NULL},
		{"H2", H2, NULL},
		{"/DT", EO_DT, NULL},
		{"/TT", EO_TT, NULL},
		{"/TEXTAREA", EO_TEXTAREA, NULL},
		{"/FONT", EO_FONT, NULL},
		{"/INPUT", EO_INPUT, NULL},
		{"/TFOOT", EO_TFOOT, NULL},
		{"/SCRIPT", EO_SCRIPT, NULL},
		{"/SELECT", EO_SELECT, NULL},
		{"/OBJECT", EO_OBJECT, NULL},
		{"/APPLET", EO_APPLET, NULL},
		{"/NOSCRIPT", EO_NOSCRIPT, NULL},
		{"/FRAMESET", EO_FRAMESET, NULL},
		{"/FIELDSET", EO_FIELDSET, NULL},
		{"/BASEFONT", EO_BASEFONT, NULL},
		{"DL", DL, NULL},
		{"DEL", DEL, NULL},
		{"/BDO", EO_BDO, NULL},
		{"LEGEND", LEGEND, NULL},
		{"/P", EO_PARA, NULL},
		{"/MAP", EO_MAP, NULL},
		{"/SUP", EO_SUP, NULL},
		{"/SAMP", EO_SAMP, NULL},
		{"UL", UL, NULL},
		{"/OPTGROUP", EO_OPTGROUP, NULL},
		{"STRONG", STRONG, NULL},
		{"NOSCRIPT", NOSCRIPT, NULL},
		{"ABBR", ABBR, NULL},
		{"TBODY", TBODY, NULL},
		{"S", S, NULL},
		{"TR", TR, NULL},
		{"OPTION", OPTION, NULL},
		{"BIG", BIG, NULL},
		{"FONT", FONT, NULL},
		{"TITLE", TITLE, html_title},
		{"TABLE", TABLE, NULL},
		{"FIELDSET", FIELDSET, NULL},
		{"FRAMESET", FRAMESET, NULL},
		{"SMALL", SMALL, NULL},
		{"TH", TH, NULL},
		{"SUB", SUB, NULL},
		{"/I", EO_ITALIC, NULL},
		{"/LI", EO_LI, NULL},
		{"PRE", PRE, NULL},
		{"LABEL", LABEL, NULL},
		{"ISINDEX", ISINDEX, NULL},
		{"B", BOLD, NULL},
		{"CAPTION", CAPTION, NULL},
		{"/DIV", EO_DIV, NULL},
		{"TD", TD, NULL},
		{"DT", DT, NULL},
		{"THEAD", THEAD, NULL},
		{"CENTER", CENTER, NULL},
		{"/EM", EO_EM, NULL},
		{"CODE", CODE, NULL},
		{"CITE", CITE, NULL},
		{"/FORM", EO_FORM, NULL},
		{"/ACRONYM", EO_ACRONYM, NULL},
		{"IFRAME", IFRAME, NULL},
		{"VAR", VAR, NULL},
		{"ADDRESS", ADDRESS, NULL},
		{"SCRIPT", SCRIPT, NULL},
		{"SELECT", SELECT, NULL},
		{"EM", EM, NULL},
		{"FORM", FORM, NULL},
		{"SUP", SUP, NULL},
		{"SAMP", SAMP, NULL},
		{"OL", OL, NULL},
		{"BASEFONT", BASEFONT, NULL},
		{"BDO", BDO, NULL},
		{"DIV", DIV, NULL},
		{"IMG", IMG, NULL},
		{"COL", COL, NULL},
		{"INS", INS, NULL},
		{"MENU", MENU, NULL},
		{"A", A, NULL},
		{"AREA", AREA, NULL},
		{"LI", LI, NULL},
		{"APPLET", APPLET, NULL},
		{"TT", TT, NULL},
		{"TEXTAREA", TEXTAREA, NULL},
		{"TFOOT", TFOOT, NULL},
		{"OBJECT", OBJECT, NULL},
		{"P", PARA, html_para},
		{"OPTGROUP", OPTGROUP, NULL},
		{"INPUT", INPUT, NULL},
		{"COLGROUP", COLGROUP, NULL},
		{"META", META, NULL},
		{"ACRONYM", ACRONYM, NULL},
		{"MAP", MAP, NULL},
		{"PARAM", PARAM, NULL},
		{"I", ITALIC, NULL}
	};

	static short lookup[] = {
		-1, 0, 1, -1, -1, -1, -1, -1,
		-1, -1, -1, -1, -1, -1, -1, -1,
		-1, -1, 2, -1, -1, -1, -1, 3,
		4, 5, -1, -230, 8, -1, 9, 10,
		-1, 11, -227, -225, -1, 16, 17, 18,
		-223, -219, 26, 27, -152, -4, 28, 29,
		-155, -3, -160, -2, -162, -2, 30, -168,
		-2, -1, -1, -1, -1, 31, -1, 32,
		-1, -1, 33, -1, -248, -1, 36, 37,
		38, -140, -2, 39, 40, 41, -1, -1,
		42, 43, -1, 44, -1, -1, 45, -1,
		46, 47, -1, -1, 48, -1, -1, -1,
		-1, 49, 50, 51, -1, -1, -1, 52,
		53, -1, 54, -1, -302, 58, 59, -300,
		-295, 64, -290, -109, -2, -1, 67, 68,
		-112, -2, -1, 69, 70, -114, -2, -119,
		-3, -1, -1, -1, 71, -1, -1, 72,
		73, -1, -1, 74, -1, -1, 75, -1,
		-1, -1, -1, -1, -1, 76, -1, 77,
		-1, 78, -1, 79, -1, 80, -1, -1,
		81, 82, -1, -400, 85, 86, -394, -392,
		-1, -345, -81, -4, 97, 98, 99, -1,
		100, 101, -1, -368, 104, -1, 105, -1,
		106, -1, 107, -1, 108, 109, 110, 111,
		112, -72, -2, -1, 113, -1, 114, 115,
		-376, -58, -2, -384, -1, 120, 121, -1,
		122, -56, -2, 123, 124, 125, -1, 126,
		127, -85, -4, -87, -2, 128, 129, -1,
		130, -91, -2, -403, -43, -2, 133, 134,
		-1, 135, -411, 138, -38, -2, 139, -1,
		-1, -1, -1, -1, -1, -1, -1, -1,
		-1, -1, 140, -1, -1, 141, -1, -1,
		-1, 142, -1, -1, -1, -437, -31, -2,
		-1, -1, -1, 145, -1, 146, -1, -1,
		-1, 147, 148, -1, -1, 149, 150, -1,
		-1, -1, -1, 151, -1, -1, -1, -1,
		152, -1, -1, -1, -1, -1, -1, -1,
		-1, 153, -1, -1, -1, -1, -1, 154,
		-1, -1, -1, 155, -1, 156, -1, 157,
		-1, -1, 158, -1, 159, -1, -1, -1,
		-1, 160, 161, 162, -1, 163, -1, -1,
		-1, -1, -1, -1, -1, -1, -1, -1,
		164, -1, -1, -1, -1, -1, -1, -1,
		-1, -1, -1, -1, -1, -1, -1, 165,
		-1, 166, -1, -1, -1, -1, -1, -1,
		-1, -1, -1, -1, -1, -1, -1, -1,
		-1, -1, -1, -1, -1, -1, 167, -1,
		-1, -1, 168, -1, -1, -1, -1, -1,
		-1, -1, -1, -1, -1, 169, -1, -1,
		170, -1, -1, -1, -1, -1, -1, -1,
		-1, -1, -1, -1, -1, -1, -1, -1,
		171, -1, 172, -1, -1, -1, -1, -1,
		-1, -1, -1, 173
	};

	if (len <= MAX_WORD_LENGTH && len >= MIN_WORD_LENGTH) {

		register int key = tag_hash(str, len);

		if (key <= MAX_HASH_VALUE && key >= 0) {

			register int index = lookup[key];

			if (index >= 0) {

				register const char *s = wordlist[index].tag;

				if ((*str == *s) && (my_strcmp(str + 1, s + 1) == 0))
					return &wordlist[index];

			} else if (index < -TOTAL_KEYWORDS) {

				register int offset = -1 - TOTAL_KEYWORDS - index;
				register html_token *wordptr = &wordlist[TOTAL_KEYWORDS + lookup[offset]];
				register html_token *wordendptr = wordptr + -lookup[offset + 1];

				while (wordptr < wordendptr) {

					register const char *s = wordptr->tag;

					if ((*str == *s) && (my_strcmp(str + 1, s + 1) == 0))
						return wordptr;

					wordptr++;
				}
			}
		}
	}

	return NULL;
}

/*
 * get_tag()
 *
 * returns the tag id for the tag string 'tag'
 */

html_token *get_tag(char *tag)
{
	html_token *token;

	if (!tag)
		return NULL;

	token = lookup_tag(tag, strlen(tag));
	if (!token)
		return NULL;

	return token;
}
