/*******************************************************************************
 *
 * location.c - Creates and parses the location box and location history.
 * 
 * Cheetah Web Browser
 * Copyright (C) 2001 Garett Spencley 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 *
 *******************************************************************************/

#include <stdio.h>

#include "location.h"
#include "htmlparser.h"
#include "buffer.h"
#include "error.h"
#include "uri.h"

GtkWidget *location;
GList *history;

/*
 * activat_location() - callback. Called when
 * location is entered 
 */

static void activate_location(void)
{
	char *URL;

	URL = gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(location)->entry));
	if(!URL) 
		return;

	open_location(URL);
}

/*
 * init_location() - creates the location box
 */

int init_location(GtkWidget *box)
{
	GtkWidget *hbox;
	GtkWidget *label;

	history = NULL;

	if(!box)
		return -1;

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(box), hbox, FALSE, TRUE, 0);

	label = gtk_label_new("Location: ");
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_widget_show(label);
	
	location = gtk_combo_new();
	gtk_box_pack_start(GTK_BOX(hbox), location, FALSE, TRUE, 0);
	gtk_widget_show(location);

	gtk_signal_connect(GTK_OBJECT(GTK_COMBO(location)->entry), 
					"activate", GTK_SIGNAL_FUNC(activate_location), 
					NULL);

	/* disables the drop down when you hit enter. */
	gtk_combo_disable_activate(GTK_COMBO(location));

	gtk_widget_show(hbox);

	return 0;
}

/* 
 * add_history_item() - add a url to the location history box
 */

void add_history_item(char *url)
{
	if(!url)
		return;
 
//	history = g_list_append(history, url);
//	gtk_combo_set_popdown_strings(GTK_COMBO(location), history);
}

/*
 * open_location() - parses a location string and opens it
 */

void open_location(char *URI)
{
	char *location_error = "DOH!\nI was unable to open the location %s\n" \
			"Please type the re-type the name and try again.";

	buffer_t *buffer = NULL;
	uri_t *uri;

	if(!URI)
		return;

	uri = create_uri_struct(URI);
	if(!uri) {
		error("Out of memory.");
		return;
	}

	if(uri->protocol == HTTP) 
		buffer = http_get_document(uri->host, uri->doc);
	else if(uri->protocol == L_FILE)
		buffer = buffer_load_file(uri->host);
	
	if(!buffer) {
		error(location_error, URI);	
		return;
	}

	add_history_item(URI);

	html_parse_document(buffer);

	dispose_buffer(buffer);
	uri_free(uri);
}
